
from enum import Enum

from qrogue.util import Logger


class Direction(Enum):
    Center = (0, 0)

    North = (0, -1)
    East = (1, 0)
    South = (0, 1)
    West = (-1, 0)
    N = North
    E = East
    S = South
    W = West

    Up = North
    Right = East
    Down = South
    Left = West
    U = Up
    R = Right
    D = Down
    L = Left

    def __init__(self, x: int, y: int):
        self.__x = x
        self.__y = y

    @staticmethod
    def from_coordinates(c_from: "Coordinate", c_to: "Coordinate") -> "Direction":
        return direction(c_from, c_to)

    @staticmethod
    def values() -> "[Direction]":
        return [Direction.North, Direction.East, Direction.South, Direction.West]

    @property
    def x(self) -> int:
        return self.__x

    @property
    def y(self) -> int:
        return self.__y

    def is_horizontal(self) -> bool:
        """

        :return: True if direction is East or West, False otherwise
        """
        return self.x != 0

    def opposite(self) -> "Direction":
        if self == Direction.North:
            return Direction.South
        elif self == Direction.East:
            return Direction.West
        elif self == Direction.South:
            return Direction.North
        elif self == Direction.West:
            return Direction.East
        else:
            return Direction.Center

    def __add__(self, other) -> "Coordinate":
        if isinstance(other, Direction):
            return Coordinate(self.x + other.x, self.y + other.y)
        elif isinstance(other, Coordinate):
            return other + self
        else:
            Logger.instance().throw(NotImplementedError(f"Adding \"{other}\" to a Coordinate is not supported!"))


class Coordinate:
    @staticmethod
    def distance(a: "Coordinate", b: "Coordinate") -> int:
        return abs(a.x - b.x) + abs(a.y - b.y)

    def __init__(self, x: int, y: int):
        self.__x = x
        self.__y = y

    @property
    def x(self) -> int:
        return self.__x

    @property
    def y(self) -> int:
        return self.__y

    def resolve(self) -> (int, int):
        return self.__x, self.__y

    def linearize(self, row_width: int) -> int:
        """
        Returns the index this Coordinate would have if it was stored in a row-wise fashion in a 1D array.

        :param row_width: width of one row of the grid stored in a corresponding 1D array
        :return:
        """
        return self.x + self.y * row_width

    def __add__(self, other) -> "Coordinate":
        if isinstance(other, Direction):
            return Coordinate(self.x + other.x, self.y + other.y)
        elif isinstance(other, Coordinate):
            return Coordinate(self.x + other.x, self.y + other.y)
        else:
            Logger.instance().throw(NotImplementedError(f"Adding \"{other}\" to a Coordinate is not supported!"))

    def __sub__(self, other) -> "Coordinate":
        if isinstance(other, Direction):
            return Coordinate(self.x - other.x, self.y - other.y)
        elif isinstance(other, Coordinate):
            return Coordinate(self.x - other.x, self.y - other.y)
        else:
            Logger.instance().throw(NotImplementedError(f"Subtracting \"{other}\" from a Coordinate is not supported!"))

    def __eq__(self, other) -> bool:
        if isinstance(other, Coordinate):
            return self.x == other.x and self.y == other.y
        return False

    def __hash__(self):
        return 61 * self.x + 51 * self.y

    def __str__(self):
        return f"({self.__x}|{self.__y})"


def direction(c_from: Coordinate, c_to: Coordinate) -> Direction:
    diff = c_to - c_from
    if diff.x == 0 and diff.y == 0:
        return Direction.Center
    if abs(diff.x) > abs(diff.y):
        if diff.x > 0:
            return Direction.East
        else:
            return Direction.West
    else:
        if diff.y > 0:
            return Direction.South
        else:
            return Direction.North


def distance(a: Coordinate, b: Coordinate) -> int:
    diff_x = a.x - b.x
    diff_y = a.y - b.y
    if diff_x < 0:
        diff_x = -diff_x
    if diff_y < 0:
        diff_y = -diff_y
    return diff_x + diff_y



import enum
from abc import ABC, abstractmethod
from typing import List

from qrogue.game.logic.collectibles import Collectible


class ControllableType(enum.Enum):
    Player = "M"
    Luke = "L"
    Test = "T"

    @staticmethod
    def values() -> "List[ControllableType]":
        return [ControllableType.Player, ControllableType.Luke, ControllableType.Test]

    def __init__(self, abbreviation: str):
        self.__abbreviation = abbreviation

    @property
    def name(self) -> str:
        return self.__abbreviation


class Controllable(ABC):
    def __init__(self, name: str):
        self.__name = name

    @property
    def name(self) -> str:
        return self.__name

    @abstractmethod
    def key_count(self) -> int:
        pass

    @abstractmethod
    def use_key(self) -> bool:
        pass

    @abstractmethod
    def get_img(self):
        pass

    @abstractmethod
    def description(self) -> str:
        pass

    @abstractmethod
    def give_collectible(self, collectible: Collectible):
        pass



from typing import List, Callable

from qrogue.game.logic.actors import Robot
from qrogue.game.world.navigation import Coordinate

from qrogue.game.world.map import Map, MapType, Room


class LevelMap(Map):
    def __init__(self, name: str, file_name: str, seed: int, rooms: List[List[Room]], robot: Robot,
                 spawn_room: Coordinate, check_achievement_callback: Callable[[str], bool],
                 trigger_event_callback: Callable[[str], None]):
        super().__init__(name, file_name, seed, rooms, robot, spawn_room, check_achievement_callback,
                         trigger_event_callback)

    def get_type(self) -> MapType:
        return MapType.Level



from abc import ABC

from qrogue.game.logic.collectibles import Collectible, CollectibleType
from qrogue.util import ShopConfig


class Pickup(Collectible, ABC):
    __DEFAULT_PRICE = 5 * ShopConfig.base_unit()

    def __init__(self, amount: int):
        super(Pickup, self).__init__(CollectibleType.Pickup)
        if amount <= 0:
            amount = 1
        self._amount = amount

    @property
    def amount(self):
        return self._amount

    def default_price(self) -> int:
        # the higher the amount the less the additional price (based on harmonic numbers)
        return int(sum([Pickup.__DEFAULT_PRICE / (i+1) for i in range(self._amount)]))


class Coin(Pickup):
    def __init__(self, amount: int = 1):
        super().__init__(amount)

    def name(self) -> str:
        return "Coin"

    def description(self) -> str:
        return "Coins are used to buy Collectibles from the Shop"

    def to_string(self):
        return f"{self.amount}$"

    def __str__(self) -> str:
        return self.to_string()


class Key(Pickup):
    def __init__(self, amount: int = 1):
        super().__init__(amount)

    def name(self) -> str:
        return "Key"

    def description(self) -> str:
        return "A Key is useful for opening locked doors."

    def to_string(self):
        return f"{self.amount} key(s)"

    def __str__(self) -> str:
        return self.to_string()


class Energy(Pickup):
    def __init__(self, amount: int = 10):
        super().__init__(amount)

    def name(self) -> str:
        return "Energy"

    def description(self) -> str:
        return "Gives back some energy to the robot so it can stay longer in a dungeon."

    def to_string(self) -> str:
        return f"Energy ({self.amount})"

    def __str__(self) -> str:
        return self.to_string()


from typing import Callable, Optional

from qrogue.game.world.dungeon_generator import ExpeditionGenerator, QrogueLevelGenerator, QrogueWorldGenerator
from qrogue.game.world.map import Map, WorldMap, MapType
from qrogue.game.world.navigation import Coordinate, Direction
from qrogue.util import CommonQuestions, Config, Logger, MyRandom, MapConfig, achievements

from qrogue.management.save_data import SaveData


__MAP_ORDER = {
    "l1v1": "l1v2",
    "l1v2": "l1v3",
    "l1v3": "l1v4",
    "l1v4": "w2",
}


def get_next(cur_level: str) -> str:
    if cur_level in __MAP_ORDER:
        return __MAP_ORDER[cur_level]
    return None


class MapManager:
    HUB_WORLD_NAME = "w0"
    __instance = None

    @staticmethod
    def instance() -> "MapManager":
        if MapManager.__instance is None:
            raise Exception("This singleton has not been initialized yet!")
        return MapManager.__instance

    def __init__(self, seed: int, show_world: Callable[[Optional[WorldMap]], None],
                 start_level: Callable[[int, Map], None]):
        if MapManager.__instance is not None:
            Logger.instance().throw(Exception("This class is a singleton!"))
        else:
            self.__rm = MyRandom(seed)
            self.__show_world = show_world
            self.__start_level = start_level
            self.__world_memory = {}    # str -> WorldMap

            generator = QrogueWorldGenerator(seed, SaveData.instance().player,
                                             SaveData.instance().achievement_manager.check_achievement,
                                             SaveData.instance().achievement_manager.add_to_achievement,
                                             self.__load_map)
            hub_world, success = generator.generate("worlds")
            if not success:
                Logger.instance().throw(RuntimeError("Unable to build world map! Please download again and make sure "
                                                     "to not edit game data."))
            self.__world_memory[self.HUB_WORLD_NAME] = hub_world
            self.__cur_map = hub_world
            self.__in_level = False

            MapManager.__instance = self

    def move_on_cur_map(self, direction: Direction) -> bool:
        return self.__cur_map.move(direction)

    @property
    def __hub_world(self) -> WorldMap:
        return self.__get_world(self.HUB_WORLD_NAME)

    def __get_world(self, level_name: str) -> WorldMap:
        if level_name[0] == "l" and level_name[1].isdigit():
            world_name = "w" + level_name[1]
            if world_name in self.__world_memory:
                return self.__world_memory[world_name]
            else:
                def trigger_achievement(name: str):
                    SaveData.instance().achievement_manager.add_to_achievement(name, 1)
                check_achievement = SaveData.instance().achievement_manager.check_achievement
                generator = QrogueWorldGenerator(self.__rm.get_seed(), SaveData.instance().player, check_achievement,
                                                 trigger_achievement, self.__load_map)
                world, success = generator.generate(world_name)
                if success:
                    self.__world_memory[world_name] = world
                    return world
                else:
                    Logger.instance().error(f"Error! Unable to build map \"{world_name}\". Returning to HubWorld")
        return self.__world_memory[self.HUB_WORLD_NAME]

    def __load_map(self, map_name: str, room: Coordinate):
        if map_name[0].lower().startswith(MapConfig.world_map_prefix()): # and map_name[1].isdigit():
            if map_name in self.__world_memory:
                self.__show_world(self.__world_memory[map_name])
            else:
                player = SaveData.instance().player
                check_achievement = SaveData.instance().achievement_manager.check_achievement
                trigger_event = self.trigger_level_event  # SaveData.instance().achievement_manager.trigger_level_event

                generator = QrogueWorldGenerator(self.__rm.get_seed(), player, check_achievement, trigger_event,
                                                 self.__load_map)
                try:
                    world, success = generator.generate(map_name)
                    if success:
                        self.__cur_map = world
                        self.__show_world(world)
                    else:
                        Logger.instance().error(f"Could not load world \"{map_name}\"!")
                except FileNotFoundError:
                    Logger.instance().error(f"Failed to open the specified world-file: {map_name}")
        elif map_name.lower().startswith(MapConfig.level_map_prefix()):
            # todo maybe levels should be able to have arbitrary names aside from "w..." or "back"?
            check_achievement = SaveData.instance().achievement_manager.check_achievement
            trigger_event = self.trigger_level_event
            generator = QrogueLevelGenerator(self.__rm.get_seed(), check_achievement, trigger_event, self.__load_map)
            try:
                level, success = generator.generate(map_name)
                if success:
                    self.__cur_map = level
                    self.__in_level = True
                    self.__start_level(self.__rm.get_seed(), self.__cur_map)
                else:
                    Logger.instance().error(f"Could not load level \"{map_name}\"!")
            except FileNotFoundError:
                Logger.instance().error(f"Failed to open the specified level-file: {map_name}")
        elif map_name.lower().startswith(MapConfig.expedition_map_prefix()):
            difficulty = int(map_name[len(MapConfig.expedition_map_prefix()):])
            seed = self.__rm.get_seed()
            robot = SaveData.instance().get_robot(0)
            check_achievement = SaveData.instance().achievement_manager.check_achievement
            trigger_event = self.trigger_level_event #SaveData.instance().achievement_manager.trigger_level_event
            generator = ExpeditionGenerator(seed, check_achievement, trigger_event, self.__load_map)
            expedition, success = generator.generate(robot)
            if success:
                self.__cur_map = expedition
                self.__in_level = True
                self.__start_level(seed, self.__cur_map)
            else:
                Logger.instance().error(f"Could not create expedition with seed = {seed}")
        elif map_name == MapConfig.back_map_string():
            if self.__in_level:
                # if we are currently in a level we return to the current world
                self.__in_level = False
                self.__show_world(self.__get_world(self.__cur_map.internal_name))
            elif self.__cur_map is self.__hub_world:
                # if we are currently in the hub-world we return to the spaceship
                self.__show_world(None)
            else:
                # if we are currently in a world we return to the hub-world
                self.__show_world(self.__hub_world)
                self.__cur_map = self.__hub_world
        else:
            Logger.instance().error(f"Invalid map to load: {map_name}")

    def __proceed(self, confirmed: bool = True):
        if confirmed:
            self.load_next()

    def trigger_level_event(self, event_id: str):
        if event_id.lower() == MapConfig.done_event_id():
            event_id = MapConfig.specific_done_event_id(self.__cur_map.internal_name)
            if self.__cur_map.get_type() is MapType.World:
                SaveData.instance().achievement_manager.finished_world(self.__cur_map.internal_name)
            elif self.__cur_map.get_type() is MapType.Level:
                SaveData.instance().achievement_manager.finished_level(self.__cur_map.internal_name)
            elif self.__cur_map.get_type() is MapType.Expedition:
                SaveData.instance().achievement_manager.add_to_achievement(achievements.CompletedExpedition, 1)
            CommonQuestions.ProceedToNextMap.ask(self.__proceed)
        SaveData.instance().achievement_manager.trigger_level_event(event_id)
        if Config.debugging():
            print("triggered event: " + event_id)

    def load_map(self, map_name: str, spawn_room: Coordinate):
        if map_name.lower() == MapConfig.next_map_string():
            self.load_next()
        else:
            self.__load_map(map_name, spawn_room)

    def load_next(self):
        next_map = get_next(self.__cur_map.internal_name)
        if next_map:
            self.load_map(next_map, None)
        else:
            world = self.__get_world(self.__cur_map.internal_name)
            self.__show_world(world)



from enum import Enum
from typing import Callable

from qrogue.util import Config, ColorConfig as CC, Logger

class _CallbackHandler:
    __show = None
    __ask = None

    @staticmethod
    def show(title: str, text: str):
        if _CallbackHandler.__show:
            _CallbackHandler.__show(title, text)
        else:
            Logger.instance().error("CommonMessages' show is None!")

    @staticmethod
    def set_show_callback(show: Callable[[str, str], None]):
        _CallbackHandler.__show = show

    @staticmethod
    def ask(text: str, callback: Callable[[bool], None]):
        if _CallbackHandler.__ask:
            _CallbackHandler.__ask(text, callback)
        else:
            Logger.instance().error("CommonMessages' ask is None!")

    @staticmethod
    def set_ask_callback(ask: Callable[[str, Callable[[bool], None]], None]):
        _CallbackHandler.__ask = ask


def set_show_callback(show: Callable[[str, str], None]):
    _CallbackHandler.set_show_callback(show)


def set_ask_callback(ask: Callable[[str, Callable[[bool], None]], None]):
    _CallbackHandler.set_ask_callback(ask)


def _locked_door() -> str:
    key = CC.highlight_object("Key")
    door = CC.highlight_object("Door")
    return f"Come back with a {key} to open the {door}."


def _one_way_door() -> str:
    one_way = CC.highlight_object("one way")
    door = CC.highlight_object("Door")
    return f"This is a {one_way} {door} that can only be opened from the other side!"


def _entangled_door() -> str:
    door = CC.highlight_object("Door")
    entangled = CC.highlight_word("entangled")
    return f"The {door} {entangled} with this one was opened. Therefore you can no longer pass this {door}."


def _tutorial_blocked() -> str:
    step = CC.highlight_word("current step")
    tutorial = CC.highlight_word("Tutorial")
    return f"You should not go there yet! Finish the {step} of the {tutorial} first."


def _not_enough_money() -> str:
    return "You cannot afford that right now. Come back when you have enough money."


def _no_space() -> str:
    circ = CC.highlight_object("Circuit")
    space = CC.highlight_word("no more space")
    gate = CC.highlight_object("Gate")
    return f"Your {circ} has {space} left. Remove a {gate} to place another one."


class CommonPopups(Enum):
    LockedDoor = ("Door is locked!", _locked_door())
    EventDoor = (Config.scientist_name(), "Hmm, I think we should complete the current task first.")
    WrongDirectionDoor = (Config.scientist_name(), "We sadly cannot access the door from this direction.")
    EntangledDoor = ("Door is entangled!", _entangled_door())
    TutorialBlocked = ("Halt!", _tutorial_blocked())
    NotEnoughMoney = ("$$$", _not_enough_money())
    NoCircuitSpace = ("Nope", _no_space())

    def __init__(self, title: str, text: str):
        self.__title = title
        self.__text = text

    def show(self):
        _CallbackHandler.show(self.__title, self.__text)


class CommonQuestions(Enum):
    __ask = None

    GoingBack = "We are not done yet. \nDo you really want to go back to the spaceship?"
    ProceedToNextMap = "Looks like we cleared this map. Shall we proceed directly to the next one?"

    def __init__(self, text: str):
        self.__text = text

    def ask(self, callback: Callable[[bool], None]):
        _CallbackHandler.ask(self.__text, callback)



# Generated from D:/Documents/pycharm_workspace/Qrogue/qrogue/dungeon_editor\QrogueWorld.g4 by ANTLR 4.9.2
from antlr4 import *

# This class defines a complete listener for a parse tree produced by QrogueWorldParser.
class QrogueWorldListener(ParseTreeListener):

    # Enter a parse tree produced by QrogueWorldParser#start.
    def enterStart(self, ctx):
        pass

    # Exit a parse tree produced by QrogueWorldParser#start.
    def exitStart(self, ctx):
        pass


    # Enter a parse tree produced by QrogueWorldParser#layout.
    def enterLayout(self, ctx):
        pass

    # Exit a parse tree produced by QrogueWorldParser#layout.
    def exitLayout(self, ctx):
        pass


    # Enter a parse tree produced by QrogueWorldParser#l_room_row.
    def enterL_room_row(self, ctx):
        pass

    # Exit a parse tree produced by QrogueWorldParser#l_room_row.
    def exitL_room_row(self, ctx):
        pass


    # Enter a parse tree produced by QrogueWorldParser#l_hallway_row.
    def enterL_hallway_row(self, ctx):
        pass

    # Exit a parse tree produced by QrogueWorldParser#l_hallway_row.
    def exitL_hallway_row(self, ctx):
        pass


    # Enter a parse tree produced by QrogueWorldParser#rooms.
    def enterRooms(self, ctx):
        pass

    # Exit a parse tree produced by QrogueWorldParser#rooms.
    def exitRooms(self, ctx):
        pass


    # Enter a parse tree produced by QrogueWorldParser#room.
    def enterRoom(self, ctx):
        pass

    # Exit a parse tree produced by QrogueWorldParser#room.
    def exitRoom(self, ctx):
        pass


    # Enter a parse tree produced by QrogueWorldParser#r_attributes.
    def enterR_attributes(self, ctx):
        pass

    # Exit a parse tree produced by QrogueWorldParser#r_attributes.
    def exitR_attributes(self, ctx):
        pass


    # Enter a parse tree produced by QrogueWorldParser#r_visibility.
    def enterR_visibility(self, ctx):
        pass

    # Exit a parse tree produced by QrogueWorldParser#r_visibility.
    def exitR_visibility(self, ctx):
        pass


    # Enter a parse tree produced by QrogueWorldParser#r_type.
    def enterR_type(self, ctx):
        pass

    # Exit a parse tree produced by QrogueWorldParser#r_type.
    def exitR_type(self, ctx):
        pass


    # Enter a parse tree produced by QrogueWorldParser#hallways.
    def enterHallways(self, ctx):
        pass

    # Exit a parse tree produced by QrogueWorldParser#hallways.
    def exitHallways(self, ctx):
        pass


    # Enter a parse tree produced by QrogueWorldParser#hallway.
    def enterHallway(self, ctx):
        pass

    # Exit a parse tree produced by QrogueWorldParser#hallway.
    def exitHallway(self, ctx):
        pass


    # Enter a parse tree produced by QrogueWorldParser#h_attributes.
    def enterH_attributes(self, ctx):
        pass

    # Exit a parse tree produced by QrogueWorldParser#h_attributes.
    def exitH_attributes(self, ctx):
        pass





from enum import IntEnum
from typing import List

from py_cui.keys import *


class Keys(IntEnum):
    MoveUp = 0
    MoveRight = MoveUp + 1
    MoveDown = MoveUp + 2
    MoveLeft = MoveUp + 3

    SelectionUp = MoveLeft + 1
    SelectionRight = SelectionUp + 1
    SelectionDown = SelectionUp + 2
    SelectionLeft = SelectionUp + 3

    PopupClose = SelectionLeft + 1
    PopupScrollUp = PopupClose + 1
    PopupScrollDown = PopupClose + 2
    PopupScrollUpFast = PopupClose + 3
    PopupScrollDownFast = PopupClose + 4

    Action = PopupScrollDownFast + 1
    Cancel = Action + 1
    Pause = Cancel + 1

    HotKey1 = Pause + 1
    HotKey2 = Pause + 2
    HotKey3 = Pause + 3
    HotKey4 = Pause + 4
    HotKey5 = Pause + 5
    HotKey6 = Pause + 6
    HotKey7 = Pause + 7
    HotKey8 = Pause + 8
    HotKey9 = Pause + 9
    HotKey0 = Pause + 10

    Render = HotKey0 + 1
    PrintScreen = Render + 1
    StopSimulator = PrintScreen + 1

    CheatInput = StopSimulator + 1
    CheatList = CheatInput + 1

    Invalid = 126

    @staticmethod
    def from_code(code: int) -> "Keys":
        num = code - 1  # since code = num + 1
        return Keys.from_index(num)

    @staticmethod
    def from_index(index: int) -> "Keys":
        i = 0
        for elem in Keys:
            if i == index:
                return elem
            i += 1
        return Keys.Invalid

    def __init__(self, num):
        self.__num = num

    @property
    def num(self) -> int:
        return self.__num

    @property
    def code(self) -> int:
        # must be a valid char
        return self.__num + 1

    def to_char(self) -> str:
        return chr(self.code)


class Controls:
    INVALID_KEY = Keys.Invalid.num

    def __init__(self):
        self.__pycui_keys = [
            # move
            [KEY_UP_ARROW, KEY_W_LOWER],
            [KEY_RIGHT_ARROW, KEY_D_LOWER],
            [KEY_DOWN_ARROW, KEY_S_LOWER],
            [KEY_LEFT_ARROW, KEY_A_LOWER],
            # select
            [KEY_UP_ARROW, KEY_W_LOWER],
            [KEY_RIGHT_ARROW, KEY_D_LOWER],
            [KEY_DOWN_ARROW, KEY_S_LOWER],
            [KEY_LEFT_ARROW, KEY_A_LOWER],
            # popups
            [KEY_ESCAPE, KEY_SPACE, KEY_ENTER],
            [KEY_UP_ARROW, KEY_W_LOWER],
            [KEY_DOWN_ARROW, KEY_S_LOWER],
            [KEY_RIGHT_ARROW, KEY_D_LOWER],
            [KEY_LEFT_ARROW, KEY_A_LOWER],

            [KEY_SPACE, KEY_ENTER],     # action
            [KEY_BACKSPACE, KEY_A_UPPER, KEY_SHIFT_LEFT],  # cancel/back
            [KEY_P_LOWER, KEY_TAB],  # pause    (Escape doesn't work here due to its special purpose for the engine)

            # special purpose hotkeys
            [KEY_1],  # Fight: Adapt (Add/Remove)
            [KEY_2],  # Fight: Commit
            [KEY_3],  # Fight: Reset
            [KEY_4],  # Fight: Items
            [KEY_5],  # Fight: Help
            [KEY_6],  # Fight: Flee
            [KEY_7],
            [KEY_8],
            [KEY_9],
            [KEY_0, 94],    # 94 = ^

            [KEY_CTRL_R],  # render screen
            [KEY_CTRL_P],   # print screen
            [KEY_ESCAPE],   # stop simulator
            [KEY_CTRL_I],   # cheat input
            [KEY_CTRL_L],   # cheat list
        ]

    def encode(self, key_pressed: int) -> Keys:
        """
        Converts a pressed key to an internal Key-representation that encodes the corresponding action
        :param key_pressed: the key that was pressed
        :return: an element of Key corresponding to the action executed by pressing key_pressed
        """
        for i in range(len(self.__pycui_keys)):
            if key_pressed == self.__pycui_keys[i]:
                return Keys.from_index(i)
        return Keys.Invalid

    def decode(self, key_code: int) -> int:
        """
        Decodes a code representation to a corresponding Keys element
        :param key_code: code representation of a Keys element
        :return: a keyboard key that can be pressed to trigger the corresponding action
        """
        if key_code == Keys.Invalid.code:
            return None
        key = Keys.from_code(key_code)
        return self.get_key(key)

    def get_keys(self, key: Keys) -> List[int]:
        return self.__pycui_keys[key.num]

    def get_key(self, key: Keys, index: int = 0):
        keys = self.get_keys(key)
        if 0 <= index < len(keys):
            return keys[index]
        return keys[0]

    @property
    def action(self) -> List[int]:
        return self.__pycui_keys[Keys.Action.num]



from typing import Callable


class Pausing:
    __instance = None

    @staticmethod
    def pause():
        if Pausing.__instance:
            Pausing.__instance.__pause_now()

    def __init__(self, callback: Callable[[], None]):
        self.__callback = callback
        Pausing.__instance = self

    def __pause_now(self):
        self.__callback()


from typing import Optional, Callable, Tuple

from qrogue.util import Config


class Message:
    __err_count = 0

    @staticmethod
    def error(text: str) -> "Message":
        Message.__err_count += 1
        return Message(f"Error_{Message.__err_count}", "Error", text, None, None)

    @staticmethod
    def create_simple(m_id: str, text: str) -> "Message":
        return Message(m_id, Config.scientist_name(), text, None, None)

    @staticmethod
    def create_with_title(m_id: str, title: str, text: str) -> "Message":
        return Message(m_id, title, text, None, None)

    @staticmethod
    def create_with_alternative(m_id: str, title: str, text: str, event: str, alternative: "Message") -> "Message":
        message = Message(m_id, title, text, event, alternative.id)
        message.resolve_message_ref(alternative)
        return message

    def __init__(self, m_id: str, title: str, text: str, event: Optional[str], alternative: Optional[str]):
        self.__m_id = m_id
        self.__title = title
        self.__text = text
        self.__alt_ref = alternative
        self.__event = event
        self.__alt_message = None

        if self.alt_message_ref is self.id:
            raise ValueError("Messages must not reference themselves as alternatives!")

    @property
    def id(self) -> str:
        return self.__m_id

    @property
    def alt_message_ref(self) -> str:
        if self.__alt_ref:
            return self.__alt_ref
        else:
            return None

    def resolve_message_ref(self, message: "Message") -> bool:
        if self.alt_message_ref == message.__m_id:
            # check for cycles
            alt = message
            length = 1
            while alt:
                if alt is self:
                    raise ValueError(f"Found cycle of length {length} in alternative messages!")
                alt = alt.__alt_message
                length += 1
            self.__alt_message = message
            return True
        return False

    def get(self, check_achievement: Callable[[str], bool]) -> Tuple[str, str]:
        if check_achievement(self.__event):
            if self.__alt_message:
                return self.__alt_message.get(check_achievement)
        else:
            return self.__title, self.__text



from typing import Callable

from qrogue.game.logic import Message
from qrogue.util import Config, PopupConfig


class Popup:
    __show_popup = None
    __check_achievement = None
    __popup_queue = []
    __cur_popup = None

    @staticmethod
    def update_popup_functions(show_popup_callback: Callable[[str, str, int], None]) -> None:
        Popup.__show_popup = show_popup_callback

    @staticmethod
    def update_check_achievement_function(check_achievement_callback: Callable[[str], bool]) -> None:
        Popup.__check_achievement = check_achievement_callback

    @staticmethod
    def on_close() -> bool:
        Popup.__cur_popup = None
        if len(Popup.__popup_queue) > 0:
            next_popup = Popup.__popup_queue.pop(0)
            next_popup.show()
            return False        # don't fully close popup
        return True     # popup no longer needed so we can fully close it

    @staticmethod
    def message(title: str, text: str, color: int = PopupConfig.default_color(), overwrite: bool = False):
        Popup(title, text, color, show=True, overwrite=overwrite)

    @staticmethod
    def scientist_says(text: str):
        Popup.message(Config.scientist_name(), text)

    @staticmethod
    def from_message(message: Message, overwrite: bool = False):
        if Popup.__check_achievement:
            ret = message.get(Popup.__check_achievement)
            if ret:
                title, text = ret
                Popup.message(title, text, overwrite=overwrite)

    def __init__(self, title: str, text: str, color: int = PopupConfig.default_color(), show: bool = True,
                 overwrite: bool = False):
        self.__title = title
        self.__text = text
        self.__color = color
        if show:
            self.show(overwrite)

    @property
    def _title(self) -> str:
        return self.__title

    @property
    def _text(self) -> str:
        return self.__text

    @property
    def _color(self) -> int:
        return self.__color

    def _base_show(self):
        Popup.__show_popup(self.__title, self.__text, self.__color)

    def _enqueue(self):
        Popup.__popup_queue.append(self)

    def show(self, overwrite: bool = False) -> None:
        if overwrite:
            Popup.__popup_queue.clear()
            Popup.__cur_popup = None
        if self.__cur_popup:
            self._enqueue()
        else:
            Popup.__cur_popup = self
            self._base_show()


class ConfirmationPopup(Popup):
    __show_popup = None

    @staticmethod
    def update_popup_function(show_popup_callback: Callable[[str, str, int, Callable[[bool], None]], None]):
        ConfirmationPopup.__show_popup = show_popup_callback

    @staticmethod
    def ask(text: str, callback: Callable[[bool], None]):
        ConfirmationPopup(Config.scientist_name(), text, callback)

    def __init__(self, title: str, text: str, callback: Callable[[bool], None],
                 color: int = PopupConfig.default_color(), show: bool = True, overwrite: bool = False):
        self.__callback = callback
        super().__init__(title, text, color, show, overwrite)

    @property
    def _callback(self) -> Callable[[bool], None]:
        return self.__callback

    def _base_show(self) -> None:
        ConfirmationPopup.__show_popup(self._title, self._text, self._color, self._callback)





import os
from setuptools import setup

README_PATH = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'README.md')
with open(README_PATH) as readme_file:
	README = readme_file.read()

setup(
	name='qrogue',  # Name of package folder
	packages=[
		'qrogue',
		'qrogue.util',
		'qrogue.game',
		'qrogue.game.logic',
		'qrogue.game.logic.collectibles',
		'qrogue.game.logic.actors',
		'qrogue.game.logic.actors.controllables',
		'qrogue.game.logic.actors.puzzles',
		'qrogue.game.world',
		'qrogue.game.world.navigation',
		'qrogue.game.world.tiles',
		'qrogue.game.world.map',
		'qrogue.game.world.dungeon_generator',
		'qrogue.game.world.dungeon_generator.dungeon_parser',
		'qrogue.game.world.dungeon_generator.world_parser',
		'qrogue.graphics',
		'qrogue.graphics.rendering',
		'qrogue.graphics.popups',
		'qrogue.graphics.widgets',
		'qrogue.management',
	],
	package_data={"data": ["qrogue/data"]},
	include_package_data=True,
	version='0.3.2.4',
	license='MIT',  # Chose a license from here: https://help.github.com/articles/licensing-a-repository
	description='Qrogue is a modernized Quantum Computing take of the classical game Rogue.',
	long_description=README,
	long_description_content_type="text/markdown",
	author='Michael Artner',
	author_email='michael.artner@jku.at',
	url='https://github.com/7Magic7Mike7/Qrogue',
	download_url='https://github.com/7Magic7Mike7/Qrogue/releases/tag/0.3.2.4',  # link to Release on GitHub
	keywords=['Quantum Computing', 'Gamification', 'Video Game', 'PyCUI'],
	install_requires=[
		'py-cui==0.1.4',
		'qiskit==0.34.2',
		'antlr4-python3-runtime==4.9.3',
	],
	classifiers=[
		'Development Status :: 3 - Alpha',  # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable"
		'Intended Audience :: Education',
		'Intended Audience :: Science/Research',
		'Topic :: Games/Entertainment :: Puzzle Games',
		'Programming Language :: Python :: 3.8',
	],
	entry_points={
		'console_scripts': ['qrogue=qrogue.qrogue:start_game'],
	}
)



from typing import Tuple

from qrogue.game.logic.actors import StateVector
from qrogue.game.logic.collectibles import Collectible

from .target import Target


class Riddle(Target):
    def __init__(self, target: StateVector, reward: Collectible, attempts: int = 1):
        super().__init__(target, reward)
        self.__attempts = attempts

    @property
    def attempts(self) -> int:
        return self.__attempts

    @property
    def is_active(self) -> bool:
        return super(Riddle, self).is_active and self.attempts > 0

    def is_reached(self, state_vector: StateVector) -> Tuple[bool, Collectible]:
        success, reward = super(Riddle, self).is_reached(state_vector)
        if not success:
            self.__attempts -= 1
        return success, reward

    def _on_reached(self):
        pass



import qrogue.util
import qrogue.game
import qrogue.graphics
import qrogue.management

from qrogue.qrogue import start_game
from qrogue.qrogue import setup_game
from qrogue.qrogue import validate_map




from abc import ABC, abstractmethod
from enum import Enum

import py_cui

from qrogue.game.logic.actors import Controllable
from qrogue.game.world.navigation import Direction


class TileCode(Enum):
    Invalid = -1    # when an error occurs, e.g. a tile at a non-existing position should be retrieved
    Debug = -2      # displays a digit for debugging
    Void = 7        # tile outside of the playable area
    Floor = 0       # simple floor tile without special meaning
    HallwayEntrance = 5     # depending on the hallway it refers to is either a Floor or Wall
    FogOfWar = 3    # tile of a place we cannot see yet

    Message = 6         # tile for displaying a popup message
    Trigger = 9         # tile that calls a function on walk, i.e. event tile
    Teleport = 91       # special trigger for teleporting between maps
    Decoration = 11     # simply displays a specified character

    Wall = 1
    Obstacle = 2
    Door = 4

    Controllable = 20
    Enemy = 30
    Boss = 40

    Collectible = 50
    Riddler = 51
    ShopKeeper = 52
    Energy = 53

    SpaceshipBlock = 70
    SpaceshipWalk = 71
    SpaceshipTrigger = 72
    OuterSpace = 73


class TileColorer:
    __color_manager = {
        TileCode.Invalid: py_cui.RED_ON_BLUE,
        TileCode.Void: py_cui.CYAN_ON_BLACK,
        TileCode.Floor: py_cui.CYAN_ON_BLACK,
        TileCode.Wall: py_cui.BLACK_ON_WHITE,
        TileCode.Obstacle: py_cui.CYAN_ON_BLACK,
        TileCode.FogOfWar: py_cui.CYAN_ON_BLACK,
        TileCode.Door: py_cui.CYAN_ON_BLACK,
        TileCode.Collectible: py_cui.CYAN_ON_BLACK,
        TileCode.Controllable: py_cui.GREEN_ON_BLACK,
        TileCode.Enemy: py_cui.RED_ON_BLACK,
        TileCode.Boss: py_cui.BLACK_ON_RED,
        TileCode.SpaceshipWalk: py_cui.BLACK_ON_WHITE,
    }

    @staticmethod
    def get_color(tile_code: TileCode) -> int:
        """

        :param tile_code: code of the Tile we want to get the default color of
        :return: integer representing one of the possible foreground-background color comibnations, None for invalid
        input
        """
        if tile_code in TileColorer.__color_manager:
            return TileColorer.__color_manager[tile_code]


class Tile(ABC):
    @staticmethod
    def _invisible_tile():
        return " "

    def __init__(self, code: TileCode):
        self.__code = code

    @property
    def code(self) -> TileCode:
        return self.__code

    @property
    def _invisible(self):
        return Tile._invisible_tile()

    @abstractmethod
    def get_img(self):
        pass

    @abstractmethod
    def is_walkable(self, direction: Direction, controllable: Controllable) -> bool:
        pass

    def __str__(self):
        return self.get_img()


class Invalid(Tile):
    def __init__(self):
        super().__init__(TileCode.Invalid)

    def get_img(self):
        return "§"

    def is_walkable(self, direction: Direction, controllable: Controllable) -> bool:
        return False


class Debug(Tile):
    def __init__(self, num: int):
        super(Debug, self).__init__(TileCode.Debug)
        self.__num = str(num)[0]

    def get_img(self) -> str:
        return self.__num

    def is_walkable(self, direction: Direction, controllable: Controllable) -> bool:
        return False


class Void(Tile):
    def __init__(self):
        super().__init__(TileCode.Floor)

    def get_img(self):
        return self._invisible

    def is_walkable(self, direction: Direction, controllable: Controllable) -> bool:
        return False


class Floor(Tile):
    @staticmethod
    def img():
        return Tile._invisible_tile()

    def __init__(self):
        super().__init__(TileCode.Floor)

    def get_img(self):
        return Floor.img()

    def is_walkable(self, direction: Direction, controllable: Controllable) -> bool:
        return True


class Wall(Tile):
    @staticmethod
    def img():
        return "#"

    def __init__(self):
        super().__init__(TileCode.Wall)

    def get_img(self):
        return Wall.img()

    def is_walkable(self, direction: Direction, controllable: Controllable) -> bool:
        return False


class Obstacle(Tile):
    def __init__(self):
        super().__init__(TileCode.Obstacle)

    def get_img(self):
        return "o"

    def is_walkable(self, direction: Direction, controllable: Controllable) -> bool:
        return False


class FogOfWar(Tile):
    def __init__(self):
        super().__init__(TileCode.Obstacle)

    def get_img(self):
        return "~"

    def is_walkable(self, direction: Direction, controllable: Controllable) -> bool:
        return True


class Decoration(Tile):
    def __init__(self, decoration: str, blocking: bool = False):
        super(Decoration, self).__init__(TileCode.Decoration)
        self.__decoration = decoration
        self.__blocking = blocking

    def get_img(self):
        return self.__decoration

    def is_walkable(self, direction: Direction, controllable: Controllable) -> bool:
        return not self.__blocking


class ControllableTile(Tile):
    def __init__(self, controllable: Controllable):
        super().__init__(TileCode.Controllable)
        self.__controllable = controllable

    def get_img(self):
        return self.__controllable.get_img()

    def is_walkable(self, direction: Direction, controllable: Controllable) -> bool:
        return True # todo check

    @property
    def controllable(self) -> Controllable:
        return self.__controllable



# exporting
from .renderable import Renderable
from .my_widgets import MyBaseWidget, Widget
from .widget_sets import MyWidgetSet
from .widget_sets import ExploreWidgetSet, NavigationWidgetSet
from .widget_sets import FightWidgetSet, BossFightWidgetSet, RiddleWidgetSet
from .widget_sets import MenuWidgetSet, PauseMenuWidgetSet, ShopWidgetSet, WorkbenchWidgetSet
from .spaceship import SpaceshipWidgetSet

# importing
# +util
# +game.logic.actors
# +game.logic.actors.puzzles
# +game.logic.collectibles
# +game.world.map
# +game.world.navigation
# +graphics.popups
# +graphics.rendering



# exporting
from .state_vector import StateVector
from .controllables import Controllable, Player, Robot
from .controllables import robot
from .puzzles import Enemy, Boss, Riddle

# importing
# +collectibles (Instruction)
# +util (is_power_2)



from typing import List, Tuple, Callable

from antlr4 import InputStream, CommonTokenStream
from antlr4.tree.Tree import TerminalNodeImpl

from qrogue.game.world.dungeon_generator import parser_util
from qrogue.game.logic import Message
from qrogue.game.logic.actors import Player
from qrogue.game.world.map import Room, MetaRoom, SpawnRoom, WorldMap
from qrogue.game.world.navigation import Coordinate, Direction
from qrogue.game.world.tiles import Door, DoorOneWayState, DoorOpenState
from qrogue.util import MapConfig, PathConfig, Logger

from .world_parser.QrogueWorldLexer import QrogueWorldLexer
from .world_parser.QrogueWorldParser import QrogueWorldParser
from .world_parser.QrogueWorldVisitor import QrogueWorldVisitor


class QrogueWorldGenerator(QrogueWorldVisitor):
    @staticmethod
    def is_spawn_room(room_id: str) -> bool:
        return room_id.lower() == 'sr'

    def __init__(self, seed: int, player: Player, check_achievement_callback: Callable[[str], bool],
                 trigger_achievement_callback: Callable[[str], None],
                 load_map_callback: Callable[[str, Coordinate], None]):
        self.__seed = seed
        self.__player = player
        self.__check_achievement = check_achievement_callback
        self.__trigger_achievement = trigger_achievement_callback
        self.__load_map = load_map_callback

        self.__hallways_by_id = {}
        self.__created_hallways = {}
        self.__hallways = {}

        self.__rooms = {}
        self.__spawn_pos = None

    def _add_hallway(self, room1: Coordinate, room2: Coordinate, door: Door):
        if door:  # for simplicity door could be null so we check it here
            if room1 in self.__hallways:
                self.__hallways[room1][room2] = door
            else:
                self.__hallways[room1] = {room2: door}
            if room2 in self.__hallways:
                self.__hallways[room2][room1] = door
            else:
                self.__hallways[room2] = {room1: door}

    def generate(self, file_name: str, in_dungeon_folder: bool = True) -> Tuple[WorldMap, bool]:
        map_data = PathConfig.read_world(file_name, in_dungeon_folder)

        input_stream = InputStream(map_data)
        lexer = QrogueWorldLexer(input_stream)
        token_stream = CommonTokenStream(lexer)
        parser = QrogueWorldParser(token_stream)
        parser.addErrorListener(parser_util.MyErrorListener())

        try:
            name, room_matrix = self.visit(parser.start())
            if name is None:
                name = file_name
        except SyntaxError as se:
            Logger.instance().error(str(se))
            return None, False

        # add empty rooms if rows don't have the same width
        max_len = 0
        for row in room_matrix:
            if len(row) > max_len:
                max_len = len(row)
        for row in room_matrix:
            if len(row) < max_len:
                row += [None] * (max_len - len(row))

        world = WorldMap(name, file_name, self.__seed, room_matrix, self.__player, self.__spawn_pos,
                         self.__check_achievement, self.__trigger_achievement)
        return world, True

    def __load_next(self):
        self.__load_map("next", None)

    ##### loading #####

    def __load_hallway(self, reference: str) -> Door:
        if reference in self.__hallways_by_id:
            return self.__hallways_by_id[reference]
        elif reference == parser_util.EMPTY_HALLWAY_CODE:
            return None
        elif reference == parser_util.DEFAULT_HALLWAY_STR:
            return Door(Direction.North)
        else:
            return Door(Direction.North)

    def __load_room(self, reference: str, x: int, y: int) -> Room:
        if reference in self.__rooms:
            if str.lower(reference) == 'sr':
                if self.__spawn_pos:
                    parser_util.warning("A second SpawnRoom was defined! Ignoring the first one "
                                        "and using this one as SpawnRoom.")
                self.__spawn_pos = Coordinate(x, y)
            room = self.__rooms[reference]
            hw_dic = parser_util.get_hallways(self.__created_hallways, self.__hallways, Coordinate(x, y))
            return room.copy(hw_dic)
        else:
            parser_util.warning(f"room_id \"{reference}\" not specified and imports not supported for worlds! "
                                "Placing an empty room instead.")
            # row.append(rooms.Placeholder.empty_room())
        return SpawnRoom(self.__load_map)

    ##### Hallway area #####

    def visitH_attributes(self, ctx: QrogueWorldParser.H_attributesContext) -> Door:
        direction = Direction.North
        door_check = None
        if ctx.OPEN_LITERAL():
            open_state = DoorOpenState.Open
        elif ctx.EVENT_LITERAL():
            if ctx.REFERENCE():
                event_id = parser_util.normalize_reference(ctx.REFERENCE().getText())

                def door_check():
                    return self.__check_achievement(event_id)
                open_state = DoorOpenState.EventLocked
            else:
                open_state = DoorOpenState.Open
                parser_util.warning("Event lock specified without an event id! Ignoring the lock and placing an open "
                                    "door instead.")
        else:
            open_state = DoorOpenState.Closed
            parser_util.warning("Invalid hallway attribute: it is neither locked nor opened nor closed!")
        return Door(direction, open_state, DoorOneWayState.NoOneWay, door_check)

    def visitHallway(self, ctx: QrogueWorldParser.HallwayContext) -> Tuple[str, Door]:
        hw_id = ctx.HALLWAY_ID().getText()
        door = self.visit(ctx.h_attributes())
        return hw_id, door

    def visitHallways(self, ctx: QrogueWorldParser.HallwaysContext) -> None:
        for hallway_ctx in ctx.hallway():
            hw_id, door = self.visit(hallway_ctx)
            self.__hallways_by_id[hw_id] = door

    ##### Room area #####

    def visitR_type(self, ctx: QrogueWorldParser.R_typeContext) -> str:
        if ctx.WORLD_LITERAL():
            return "W"
        elif ctx.LEVEL_LITERAL():
            return "L"
        else:
            raise ValueError(f"Invalid r_type: {ctx.getText()}")

    def visitR_visibility(self, ctx: QrogueWorldParser.R_visibilityContext) -> Tuple[bool, bool]:
        visible = False
        foggy = False
        if ctx.VISIBLE_LITERAL():
            visible = True
        elif ctx.FOGGY_LITERAL():
            foggy = True
        return visible, foggy

    def visitR_attributes(self, ctx: QrogueWorldParser.R_attributesContext) \
            -> Tuple[Tuple[bool, bool], bool, int, Direction]:
        visibility = self.visit(ctx.r_visibility())
        rtype = self.visit(ctx.r_type())
        num = 0
        for digit in ctx.DIGIT():
            num *= 10
            d = int(digit.getText())
            num += d
        direction = parser_util.direction_from_string(ctx.DIRECTION().getText())
        return visibility, rtype, num, direction

    def visitRoom(self, ctx: QrogueWorldParser.RoomContext) -> Tuple[str, MetaRoom]:
        room_id = ctx.ROOM_ID().getText()
        msg = ctx.TEXT().getText()[1:-1]  # strip encapsulating \"
        level_to_load = parser_util.normalize_reference(ctx.REFERENCE().getText())
        visibility, m_type, num, orientation = self.visit(ctx.r_attributes())

        alt_message = Message.create_with_title(f"load{room_id}Done", f"{level_to_load} - done", msg)
        # the (internal) level name is also the name of the event that describes whether the level was completed or not
        message = Message.create_with_alternative(f"load{room_id}", level_to_load, msg, level_to_load, alt_message)
        # hallways will be added later
        if self.is_spawn_room(room_id):
            room = MetaRoom(self.__load_map, orientation, message, level_to_load, m_type, num, is_spawn=True)
        else:
            room = MetaRoom(self.__load_map, orientation, message, level_to_load, m_type, num)
        visible, foggy = visibility
        if visible:
            room.make_visible()
        elif foggy:
            room.in_sight()
        return room_id, room

    def visitRooms(self, ctx: QrogueWorldParser.RoomsContext):
        for room_ctx in ctx.room():
            room_id, room = self.visit(room_ctx)
            self.__rooms[room_id] = room

    ##### Layout area #####

    def visitLayout(self, ctx: QrogueWorldParser.LayoutContext) -> List[List[Room]]:
        # first setup all hallway connections
        for y, hw_row in enumerate(ctx.l_hallway_row()):
            self.__visitL_hallway_row(hw_row, y)

        room_matrix = []
        for y in range(MapConfig.max_height()):
            row_ctx = ctx.l_room_row(y)
            if row_ctx:
                room_matrix.append(self.__visitL_room_row(row_ctx, y))
            else:
                break
        if ctx.l_room_row(MapConfig.max_height()):
            parser_util.warning(
                f"Too much room rows specified. Only maps of size ({MapConfig.max_width()}, "
                f"{MapConfig.max_height()}) supported. Ignoring over-specified rows.")

        return room_matrix

    def __hallway_handling(self, ctx_children: List[TerminalNodeImpl], y: int, direction: Direction):
        x = 0
        for child in ctx_children:
            if parser_util.check_for_overspecified_columns(x, child.symbol.type, QrogueWorldParser.VERTICAL_SEPARATOR):
                parser_util.warning(
                    f"Too much room columns specified. Only maps of size ({MapConfig.max_width()}, "
                    f"{MapConfig.max_height()}) supported. Ignoring over-specified columns.")
                break
            if child.symbol.type == QrogueWorldParser.HALLWAY_ID:
                hw_id = child.symbol.text
                origin = Coordinate(x, y)
                self._add_hallway(origin, origin + direction, self.__load_hallway(hw_id))
                x += 1
            elif child.symbol.type == QrogueWorldParser.EMPTY_HALLWAY:
                x += 1

    def __visitL_hallway_row(self, ctx: QrogueWorldParser.L_hallway_rowContext, y: int) -> None:
        self.__hallway_handling(ctx.children, y, Direction.South)  # connect downwards to the next room row

    def __visitL_room_row(self, ctx: QrogueWorldParser.L_room_rowContext, y: int) -> List[Room]:
        self.__hallway_handling(ctx.children, y, Direction.East)  # connect to the right to the next room

        row = []
        x = 0
        for child in ctx.children:
            if parser_util.check_for_overspecified_columns(x, child.symbol.type, QrogueWorldParser.VERTICAL_SEPARATOR):
                parser_util.warning(f"Too much room columns specified. Only maps of size ({MapConfig.max_width()}, "
                                    f"{MapConfig.max_height()}) supported. Ignoring over-specified columns.")
                break

            if child.symbol.type == QrogueWorldParser.ROOM_ID:
                room_id = child.symbol.text  # todo make it illegal to have the same room_id twice?
                row.append(self.__load_room(room_id, x, y))
                x += 1
            elif child.symbol.type == QrogueWorldParser.EMPTY_ROOM:
                row.append(None)
                x += 1
        return row

    ##### Start area #####

    def visitStart(self, ctx: QrogueWorldParser.StartContext) -> Tuple[str, List[List[Room]]]:
        if ctx.NAME():
            name = ctx.TEXT().getText()[1:-1]
        else:
            name = None

        # prepare hallways first because they are standalone
        self.visit(ctx.hallways())
        # next prepare the rooms because they reference hallways
        self.visit(ctx.rooms())

        # for the last step we retrieve the room matrix from layout
        return name, self.visit(ctx.layout())



from typing import Dict

from antlr4.error.ErrorListener import ErrorListener

from qrogue.game.world.map import Hallway
from qrogue.game.world.navigation import Coordinate, Direction
from qrogue.util import MapConfig, Logger, Config

DEFAULT_HALLWAY_STR = "=="
TEMPLATE_PREFIX = "_"
EMPTY_ROOM_CODE = "_a"
EMPTY_HALLWAY_CODE = "_0"

OBSTACLE_TILE = "o"
PLACE_HOLDER_TILE = "_"
COLLECTIBLE_TILE = "c"
TRIGGER_TILE = "t"
MESSAGE_TILE = "m"
ENERGY_TILE = "e"
RIDDLER_TILE = "r"
SHOP_KEEPER_TILE = "$"
FLOOR_TILE = " "


def warning(text: str):
    Logger.instance().println(f"Warning: {text}")
    if Config.debugging():
        print("Warning", text)


def check_for_overspecified_columns(x: int, symbol_type, ref_type):
    return x == MapConfig.max_width() and symbol_type != ref_type or x > MapConfig.max_width()


def normalize_reference(reference: str) -> str:
    if reference[0] == '*':
        return reference[1:].lower()
    else:
        return reference.lower()


def direction_from_string(dir_str: str) -> Direction:
    direction = None
    if dir_str == "North":
        direction = Direction.North
    elif dir_str == "East":
        direction = Direction.East
    elif dir_str == "South":
        direction = Direction.South
    elif dir_str == "West":
        direction = Direction.West
    return direction


def get_hallways(hallway_dictionary: Dict[Coordinate, Dict[Direction, Hallway]], hallways, pos: Coordinate) \
        -> Dict[Direction, Hallway]:
    if pos in hallways:
        hallways = hallways[pos]
        if hallways:
            room_hallways = {
                Direction.North: None, Direction.East: None, Direction.South: None, Direction.West: None,
            }
            for neighbor in hallways:
                direction = Direction.from_coordinates(pos, neighbor)
                opposite = direction.opposite()
                # get hallway from neighbor if it exists, otherwise create it
                if neighbor in hallway_dictionary and opposite in hallway_dictionary[neighbor]:
                    hallway = hallway_dictionary[neighbor][opposite]
                else:
                    door = hallways[neighbor]
                    if door.direction not in [direction, direction.opposite()]:
                        if door.is_one_way:
                            warning("Found one way door with invalid direction! Horizontal hallways can only "
                                         "have East or West doors and vertical ones only North and South doors! "
                                         "Removing the one way aspect and setting the direction to a valid one.")
                            door = door.copy(direction)
                        else:
                            door = door.copy(direction)
                    hallway = Hallway(door)
                    if neighbor in hallway_dictionary:
                        hallway_dictionary[neighbor][opposite] = hallway
                    else:
                        hallway_dictionary[neighbor] = {opposite: hallway}

                # store the hallway so the neighbors can find it if necessary
                if pos not in hallway_dictionary:
                    hallway_dictionary[pos] = {}
                hallway_dictionary[pos][direction] = hallway
                room_hallways[direction] = hallway
            return room_hallways
    return None


class MyErrorListener(ErrorListener):
    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        #print(f"Syntax Error: \"{offendingSymbol}\" at line {line}, column {column} - {msg}")
        raise SyntaxError(msg)

    def reportAmbiguity(self, recognizer, dfa, startIndex, stopIndex, exact, ambigAlts, configs):
        Config.debug_print("Ambiguity")

    def reportAttemptingFullContext(self, recognizer, dfa, startIndex, stopIndex, conflictingAlts, configs):
        Config.debug_print("Attempting full context")

    def reportContextSensitivity(self, recognizer, dfa, startIndex, stopIndex, prediction, configs):
        Config.debug_print("Context sensitivity")





# exporting
from .generator import DungeonGenerator
from .random_generator import ExpeditionGenerator
from qrogue.game.world.dungeon_generator.QrogueLevelGenerator import QrogueLevelGenerator
from qrogue.game.world.dungeon_generator.QrogueWorldGenerator import QrogueWorldGenerator

# importing
# +util
# +logic
# +logic.actors
# +logic.actors.controllables
# +logic.actors.puzzles
# +logic.collectibles
# +map
# +navigation
# +tiles



import random
from typing import Callable, List

from qrogue.game.logic.actors import Robot, Controllable, Player
from qrogue.game.world.navigation import Direction, Coordinate
from qrogue.game.world.tiles import Tile, TileCode, WalkTriggerTile
from qrogue.util import MyRandom, Config

ascii_spaceship = \
    r"                                                                                              " + "\n" \
    r"                                    X---------------X                                         " + "\n" \
    r"                                   /ööööööööööööööööö)>                                       " + "\n" \
    r"                                  /öööööööööööööööööö)>                           /)          " + "\n" \
    r"                                 /ööööööööööööööööööö)>                          /ö|          " + "\n" \
    r"                                |öööööööööööööööööööö)>                         /öö|          " + "\n" \
    r"                                |öööööööööööööööööööö)>                        /ööö|          " + "\n" \
    r"                                |ööööööööööööööööööö/                         /öööö|          " + "\n" \
    r"                               /ööööööööööööööööööö/                         /ööööö|          " + "\n" \
    r"                              /ööööööööööööööööööö/                         /öööööö|          " + "\n" \
    r"                             /ööööööööööööööööööö/                         /ööööööö|          " + "\n" \
    r"              X-------------Xööööööööööööööööööö(                         /öööööööö|          " + "\n" \
    r"             /ööööööööööööööööööööööööööööööööööö\                       /ööööööööö|          " + "\n" \
    r"            /ööööööööööööööööööööööööööööööööööööö\                     /öööööööööö|          " + "\n" \
    r"           /ööööööööWöööööööööööööööööööööööööööööö--------------------Xööööööööööö|          " + "\n" \
    r"          |öööööööööööööööööööööööööööööööööööööööööööööööööööööööööööööööööööööööö|          " + "\n" \
    r"          |ööSööööööööööööööööööööööööööööööööööööööööööööööööööööööööööööööööööööö|          " + "\n" \
    r"          |öööööööööööööööööööööööööööööööööööööööööööööööööööööööööööööööööööööööö|          " + "\n" \
    r"           \ööööööööööööööööööööööööööööööööööööööö--------------------Xööööööööööö|          " + "\n" \
    r"            \ööööööööööööööööööööööööööööööööööööö/                     \öööööööööö|          " + "\n" \
    r"             \ööööööööööölööööööööööööööööööööööö/                       \ööööööööö|          " + "\n" \
    r"              X-------------Xööööööööööööööööööö(                         \öööööööö|          " + "\n" \
    r"                             \ööööööööööööööööööö\                         \ööööööö|          " + "\n" \
    r"                              \ööööööööööööööööööö\                         \öööööö|          " + "\n" \
    r"                               \ööööööööööööööööööö\                         \ööööö|          " + "\n" \
    r"                                |ööööööööööööööööööö\                         \öööö|          " + "\n" \
    r"                                |öööööööööööööööööööö)>                        \ööö|          " + "\n" \
    r"                                |öööööööööööööööööööö)>                         \öö|          " + "\n" \
    r"                                 \ööööööööööööööööööö)>                          \ö|          " + "\n" \
    r"                                  \öööööööööööööööööö)>                           \)          " + "\n" \
    r"                                   \ööööööööGöööööööö)>                                       " + "\n" \
    r"                                    X---------------X                                         " + "\n" \
    r"                                                                                              " + "\n" \



__rm = random.Random()
def spaceship_random() -> float:
    return __rm.random()


class SpaceshipWallTile(Tile):
    MAP_INVISIBLE_PRESENTATION = "X"

    def __init__(self, character):
        super(SpaceshipWallTile, self).__init__(TileCode.SpaceshipBlock)
        self.__img = character

    def get_img(self):
        return self.__img

    def is_walkable(self, direction: Direction, robot: Robot) -> bool:
        return False

    def __str__(self):
        return "W"


class SpaceshipFreeWalkTile(Tile):
    MAP_REPRESENTATION = "ö"

    def __init__(self):
        super(SpaceshipFreeWalkTile, self).__init__(TileCode.SpaceshipWalk)

    def get_img(self):
        #return ColorConfig.colorize(ColorCodes.SPACESHIP_FLOOR, self._invisible)
        return "."#self._invisible

    def is_walkable(self, direction: Direction, robot: Robot) -> bool:
        return True

    def __str__(self):
        return "f"


class SpaceshipTriggerTile(WalkTriggerTile):
    MAP_START_REPRESENTATION = "S"
    MAP_WORKBENCH_REPRESENTATION = "W"
    MAP_GATE_LIBRARY_REPRESENTATION = "G"
    START_TEST_LEVEL = "l"

    def __init__(self, character: str, callback: Callable[[Direction, Robot], None]):
        super().__init__(TileCode.SpaceshipTrigger)
        self.__img = character
        self.__callback = callback

    def _on_walk(self, direction: Direction, robot: Robot) -> None:
        self.__callback(direction, robot)

    def get_img(self):
        return self.__img


class OuterSpaceTile(Tile):
    MAP_REPRESENTATION = " "
    __STAR_BIRTH_CHANCE = 0.0001
    __STAR_DIE_CHANCE = 0.01

    def __init__(self):
        super(OuterSpaceTile, self).__init__(TileCode.OuterSpace)
        self.__is_star = spaceship_random() < 0.007

    def get_img(self):
        if self.__is_star:
            if spaceship_random() < OuterSpaceTile.__STAR_DIE_CHANCE:
                self.__is_star = False
            return "*"
        else:
            if spaceship_random() < OuterSpaceTile.__STAR_BIRTH_CHANCE:
                self.__is_star = True
            return self._invisible

    def is_walkable(self, direction: Direction, robot: Robot) -> bool:
        return False

    def __str__(self):
        return "o"


class SpaceshipMap:
    WIDTH = ascii_spaceship.index("\n")
    HEIGHT = ascii_spaceship.count("\n")
    SPAWN_POS = Coordinate(x=25, y=16)

    def __init__(self, seed: int, player: Player, open_world_view: Callable[[Direction, Controllable], None],
                 use_workbench: Callable[[Direction, Controllable], None],
                 start_test_level: Callable[[Direction, Controllable], None]):
        self.__rm = MyRandom(seed)
        self.__player = player
        self.__open_world_view = open_world_view
        self.__use_workbench = use_workbench
        self.__start_test_level = start_test_level
        self.__tiles = []
        row = []
        for character in ascii_spaceship:
            if character == "\n":
                self.__tiles.append(row)
                row = []
            else:
                tile = self.__ascii_to_tile(character)
                row.append(tile)

        self.__player_pos = SpaceshipMap.SPAWN_POS

        # self.widget.activate_custom_draw()

    @property
    def player_pos(self) -> Coordinate:
        return self.__player_pos

    def __ascii_to_tile(self, character: str) -> Tile:
        if character == SpaceshipFreeWalkTile.MAP_REPRESENTATION:
            tile = SpaceshipFreeWalkTile()
        elif character == OuterSpaceTile.MAP_REPRESENTATION:
            tile = OuterSpaceTile()
        elif character == SpaceshipWallTile.MAP_INVISIBLE_PRESENTATION:
            tile = SpaceshipWallTile(" ")
        elif character == SpaceshipTriggerTile.MAP_START_REPRESENTATION:
            tile = SpaceshipTriggerTile(character, self.__open_world_view)
        elif character == SpaceshipTriggerTile.MAP_WORKBENCH_REPRESENTATION:
            tile = SpaceshipTriggerTile(character, self.__use_workbench)
        # elif character == SpaceshipTriggerTile.MAP_GATE_LIBRARY_REPRESENTATION:
        #    tile = SpaceshipTriggerTile(character, self.open_gate_library)
        elif character == SpaceshipTriggerTile.START_TEST_LEVEL:
            if Config.debugging():
                tile = SpaceshipTriggerTile(character, self.__start_test_level)
            else:
                tile = SpaceshipFreeWalkTile()
        else:
            tile = SpaceshipWallTile(character)
        return tile

    @property
    def width(self) -> int:
        return SpaceshipMap.WIDTH

    @property
    def height(self) -> int:
        return SpaceshipMap.HEIGHT

    def move(self, direction: Direction) -> bool:
        """
        Tries to move the player in the given Direction.
        :param direction: in which direction the player should move
        :return: True if the player was able to move, False otherwise
        """
        new_pos = self.__player_pos + direction
        if new_pos.y < 0 or self.height <= new_pos.y or \
                new_pos.x < 0 or self.width <= new_pos.x:
            return False

        tile = self.__tiles[new_pos.y][new_pos.x]
        if tile.is_walkable(direction, self.__player):
            if isinstance(tile, WalkTriggerTile):
                tile.trigger(direction, self.__player, self.__trigger_event)
            self.__player_pos = new_pos
            return True
        else:
            return False

    def __trigger_event(self, event_id: str):
        # todo check if we really don't need this in the Spaceship
        pass

    def get_row_strings(self) -> List[str]:
        rows = []
        for row in self.__tiles:
            cur_row = ""
            for tile in row:
                cur_row += tile.get_img()
            rows.append(cur_row)
        return rows



from abc import ABC, abstractmethod
from typing import Tuple

from qrogue.game.world.map import Map
from qrogue.util import MapConfig


class DungeonGenerator(ABC):
    WIDTH = MapConfig.max_width()
    HEIGHT = MapConfig.max_height()

    def __init__(self, seed: int, width: int = WIDTH, height: int = HEIGHT):
        self.__seed = seed
        self._width = width
        self._height = height

    @property
    def seed(self) -> int:
        return self.__seed

    @property
    def width(self) -> int:
        return self._width

    @property
    def height(self) -> int:
        return self._height

    @abstractmethod
    def generate(self, data) -> Tuple[Map, bool]:
        pass



from typing import Callable, List

import py_cui

from qrogue.game.world.map import SpaceshipMap
from qrogue.game.world.navigation import Coordinate, Direction
from qrogue.graphics.rendering import ColorRules
from qrogue.util import Controls

from qrogue.graphics.widgets import MyBaseWidget, MyWidgetSet, Renderable, Widget


class SpaceshipWidget(Widget):
    def __init__(self, widget: MyBaseWidget):
        super().__init__(widget)
        self.__spaceship_map = None
        ColorRules.apply_spaceship_rules(widget)

    def set_data(self, data: SpaceshipMap) -> None:
        self.__spaceship_map = data
        self.render()

    def render(self) -> None:
        if self.__spaceship_map:
            rows = self.__spaceship_map.get_row_strings()
            # add player on top
            x = self.__spaceship_map.player_pos.x
            y = self.__spaceship_map.player_pos.y
            rows[y] = rows[y][0:x] + "M" + rows[y][x + 1:]  # todo fix player-img
            self.widget.set_title("\n".join(rows))

    def render_reset(self) -> None:
        pass

    def move(self, direction: Direction) -> bool:
        """
        Tries to move the player in the given Direction.
        :param direction: in which direction the player should move
        :return: True if the player was able to move, False otherwise
        """
        return self.__spaceship_map.move(direction)


class SpaceshipWidgetSet(MyWidgetSet):
    def __init__(self, controls: Controls, logger, root: py_cui.PyCUI,
                 base_render_callback: Callable[[List[Renderable]], None]):
        super().__init__(controls, logger, root, base_render_callback)

    def init_widgets(self, controls: Controls) -> None:
        spaceship = self.add_block_label("Dynamic Spaceship", 0, 0, MyWidgetSet.NUM_OF_ROWS, MyWidgetSet.NUM_OF_COLS,
                                         center=True)
        self.__spaceship = SpaceshipWidget(spaceship)

    def get_widget_list(self) -> "list of Widgets":
        return [
            self.__spaceship,
        ]

    def get_main_widget(self) -> MyBaseWidget:
        return self.__spaceship.widget

    def reset(self) -> None:
        pass

    def set_data(self, data: SpaceshipMap):
        self.__spaceship.set_data(data)

    def move_up(self):
        if self.__spaceship.move(Direction.Up):
            self.render()

    def move_right(self):
        if self.__spaceship.move(Direction.Right):
            self.render()

    def move_down(self):
        if self.__spaceship.move(Direction.Down):
            self.render()

    def move_left(self):
        if self.__spaceship.move(Direction.Left):
            self.render()



# exporting
import qrogue.game.world.navigation
import qrogue.game.world.tiles
import qrogue.game.world.map
import qrogue.game.world.dungeon_generator

# importing
# +util
# +logic
# +logic.actors
# +logic.actors.controllables
# +logic.actors.puzzles
# +logic.collectibles




from qrogue.game.logic.collectibles import Collectible, CollectibleType


class Qubit(Collectible):
    __MIN_NUM = 1
    __MAX_NUM = 3

    def __init__(self, num: int = 1):
        if self.__MIN_NUM <= num <= self.__MAX_NUM:
            super().__init__(CollectibleType.Qubit)
            self.__num = num
        else:
            raise ValueError(f"Error trying to add to many qubits: {num}! Only between {self.__MIN_NUM} and "
                             f"{self.__MAX_NUM} allowed at once.")

    @property
    def additional_qubits(self) -> int:
        return self.__num

    def name(self) -> str:
        return "Qubit"

    def description(self) -> str:
        return "Qubits are the essence of any circuit you create."

    def default_price(self) -> int:
        return 100  # currently the plan is to not have it in shops

    def to_string(self) -> str:
        return f"Qubit ({self.additional_qubits})"



from datetime import datetime

import py_cui.debug

from qrogue.util.config import PathConfig, Config


class Logger(py_cui.debug.PyCUILogger):
    __BUFFER_SIZE = 2048
    __instance = None

    @staticmethod
    def instance() -> "Logger":
        if Logger.__instance is None:
            raise Exception("This singleton has not been initialized yet!")
        return Logger.__instance

    def __init__(self, seed: int):
        super().__init__("Qrogue-Logger")
        if Logger.__instance is not None:
            self.throw(Exception("This class is a singleton!"))
        else:
            self.__text = ""
            self.__message_popup = None
            self.__error_popup = None
            self.__save_file = PathConfig.new_log_file(seed)
            self.__buffer_size = 0
            self.__buffer = [Config.get_log_head(seed)]
            Logger.__instance = self

    def set_popup(self, message_popup_function: "void(str, str)", error_popup_function: "void(str, str)") -> None:
        self.__message_popup = message_popup_function
        self.__error_popup = error_popup_function

    def __write(self, text) -> None:
        self.__buffer.append(text)
        self.__buffer_size += len(text)
        if self.__buffer_size >= Logger.__BUFFER_SIZE:
            self.flush()

    def info(self, message, **kwargs) -> None:
        time_str = datetime.now().strftime("%H-%M-%S")
        self.__write(f"{time_str}: {message}")

    def debug(self, msg: str, *args, **kwargs) -> None:
        if Config.debugging():
            self.info(f"{{DEBUG}} |{msg}")

    def show_error(self, message) -> None:
        self.__error_popup("ERROR", str(message))

    def error(self, message, show: bool = True, **kwargs) -> None:
        if show:
            self.__error_popup("ERROR", str(message))
        highlighting = "\n----------------------------------\n"
        self.info(f"{highlighting}ERROR |{message}{highlighting}")

    def throw(self, error) -> None:
        print(error)
        self.__write(f"[ERROR] {error}")
        self.flush()
        raise error

    def print(self, message: str, clear: bool = False) -> None:
        print(message)
        if clear:
            self.__text = message
        else:
            self.__text += message
        self.__message_popup("Logger", self.__text)

    def println(self, message: str = "", clear: bool = False) -> None:
        self.print(f"{message}\n", clear)

    def print_list(self, list=[], delimiter: str = ", ") -> None:
        sb = "["
        for elem in list:
            if elem is not None:
                sb += str(elem)
            sb += delimiter
        sb += "]"
        self.print(sb)

    def clear(self) -> None:
        self.__text = ""

    def flush(self) -> None:
        if self.__buffer_size > 0:
            text = ""
            for log in self.__buffer:
                text += log + "\n"
            PathConfig.write(self.__save_file, text, may_exist=True, append=True)
            self.__buffer = []
            self.__buffer_size = 0
